import typing as tp
from types import GenericAlias

from pybyntic.buffer import Buffer


class DataParser:
    def __init__(self, data: bytes) -> None:
        self.buf = Buffer(data)

    def read(self, field: tp.Any) -> tp.Any:
        if isinstance(field, GenericAlias) and field.__origin__ is list:
            list_len = self.buf.read_varint()
            value = []
            for i in range(list_len):
                value.append(self.read(field.__args__[0]))
            return value
        value = field.read(self.buf)
        return value



from typing import get_args, get_origin

from pybyntic.buffer import Buffer


class DataDumper:
    def __init__(self) -> None:
        self.buf = Buffer(b"")

    def dump(self, field_type, value):
        def _dump(ft, v):
            origin = get_origin(ft)

            # Handle list[...] (works for list[T], list[list[T]], etc.)
            if origin is list:
                (elem_t,) = get_args(ft)
                self.buf.write_varint(len(v))
                for item in v:
                    _dump(elem_t, item)
                return

            # Leaf type: delegate to the type's writer
            ft.write(self.buf, v)

        _dump(field_type, value)

    def get_value(self):
        return bytes(self.buf.buffer)



"""
PyByntic extends Pydantic with binary-typed fields and automatic byte-level serialization.
Define models using familiar Pydantic syntax and turn them into compact binary payloads
with full control over layout and numeric precision.
"""
__version__ = "0.1.4"

from pybyntic.annotated_base_model import AnnotatedBaseModel
from pybyntic.types import *

__all__ = ["AnnotatedBaseModel"]



import json
from datetime import date, datetime, timedelta, timezone
from typing import Any, Generic, Protocol, Type, TypeVar

from pybyntic.buffer import Buffer

ZERO_UTC = datetime(1970, 1, 1, tzinfo=timezone.utc)


class TypeProto(Protocol):
    @classmethod
    def read(cls, buf: Buffer) -> Any:
        ...

    @classmethod
    def write(cls, buf: Buffer, value: Any) -> None:
        ...


class Skip:
    @classmethod
    def read(cls, _: Buffer) -> None:
        return None

    @classmethod
    def write(cls, buf: Buffer, value: Any):
        pass


class Bool:
    pytype = bool

    @classmethod
    def read(cls, buf: Buffer) -> bool:
        return bool(buf.read_formated("b"))

    @classmethod
    def write(cls, buf: Buffer, value: bool):
        buf.write_formated("b", 1 if value else 0)


class FmtType:
    fmt: str = "b"

    @classmethod
    def read(cls, buf: Buffer) -> int | float:
        return buf.read_formated(cls.fmt)

    @classmethod
    def write(cls, buf: Buffer, value: int | float):
        buf.write_formated(cls.fmt, value)


class Int8(FmtType):
    fmt = "b"


class Int16(FmtType):
    fmt = "h"


class Int32(FmtType):
    fmt = "i"


class Int64(FmtType):
    fmt = "q"


class UInt8(FmtType):
    fmt = "B"


class UInt16(FmtType):
    fmt = "H"


class UInt32(FmtType):
    fmt = "I"


class UInt64(FmtType):
    fmt = "Q"


class Float32(FmtType):
    fmt = "f"


class Float64(FmtType):
    fmt = "d"


class UInt128:
    @classmethod
    def read(cls, buf: Buffer) -> int:
        hi = buf.read_formated("Q")
        lo = buf.read_formated("Q")
        return (hi << 64) + lo

    @classmethod
    def write(cls, buf: Buffer, value: int):
        hi = value >> 64
        lo = value & 0xFFFFFFFFFFFFFFFF
        buf.write_formated("Q", hi)
        buf.write_formated("Q", lo)


class String:
    """
    A variable-length string prefixed with its length as a VarInt.
    UTF-8 encoding is used.
    """

    @classmethod
    def read(cls, buf: Buffer) -> str:
        length = buf.read_varint()
        data = buf.read_fixed_str(length)
        return data

    @classmethod
    def write(cls, buf: Buffer, value: str):
        data = value.encode("utf-8")
        buf.write_varint(len(data))
        buf.write_bytes(data)


class FixedString:
    """
    A fixed-length string of N bytes (not characters).
    """

    def __init__(self, length: int, encoding: str = "utf-8"):
        assert length >= 1  # FixedString must have a length of at least 1

        self.length = length
        self.encoding = encoding

    @classmethod
    def __class_getitem__(cls, args: int | tuple[int, str]):
        if isinstance(args, tuple):
            length, encoding = args
            return cls(length, encoding)
        else:
            return cls(args)

    def read(self, buf: Buffer) -> str:
        data = buf.read_fixed_str(self.length, self.encoding)
        return data

    def write(self, buf: Buffer, value: str):
        data = value.encode(self.encoding)
        if len(data) > self.length:
            data = data[: self.length]
        elif len(data) < self.length:
            data += b"\x00" * (self.length - len(data))
        buf.write_bytes(data)


class StringJson:
    @classmethod
    def read(cls, buf: Buffer) -> dict[str, Any]:
        length = buf.read_varint()
        data = buf.read_fixed_str(length)
        return json.loads(data if data else "{}")

    @classmethod
    def write(cls, buf: Buffer, value: dict[str, Any]):
        data = json.dumps(value).encode()
        buf.write_varint(len(data))
        buf.write_bytes(data)


class DateTime32:
    @classmethod
    def read(cls, buf: Buffer) -> datetime:
        if not buf.is_buffer_readable():
            return ZERO_UTC

        return datetime.fromtimestamp(buf.read_formated("I"), tz=timezone.utc)

    @classmethod
    def write(cls, buf: Buffer, value: datetime):
        offset = value.utcoffset()
        if offset is not None and offset.total_seconds() > 0:
            raise ValueError("Use DateTime32TZ for timezone-aware datetime")
        value = value.astimezone(timezone.utc)
        ts = int(value.timestamp())
        if ts < 0 or ts > 2 ** 32 - 1:
            raise ValueError("DateTime value out of range for DateTime32, use DateTime64 instead")
        UInt32.write(buf, ts)


class DateTime32TZ:
    @classmethod
    def read(cls, buf: Buffer) -> datetime:
        timestamp = UInt32.read(buf)
        offset_minutes = Int16.read(buf)
        tz = timezone(timedelta(minutes=offset_minutes))
        return datetime.fromtimestamp(timestamp, tz=tz)

    @classmethod
    def write(cls, buf: Buffer, value: datetime):
        offset_td = value.utcoffset()
        if offset_td is None:
            raise ValueError("DateTime32TZ requires timezone-aware datetime")
        offset = int(offset_td.total_seconds() / 60)
        ts = int(value.timestamp())
        if ts < 0 or ts > 2**32 - 1:
            raise ValueError("DateTime value out of range for DateTime32, use DateTime64 instead")
        UInt32.write(buf, ts)
        Int16.write(buf, offset)


class DateTime64:
    precision = 3

    def __init__(self, precision):
        self.precision = precision

    @classmethod
    def __class_getitem__(cls, precision=3):
        return cls(precision)

    def read(self, buf: Buffer) -> datetime:
        ticks = Int64.read(buf)

        divisor = 10**self.precision

        timestamp = ticks / divisor

        return datetime.fromtimestamp(timestamp, tz=timezone.utc)

    def write(self, buf: Buffer, value: datetime):
        offset = value.utcoffset()
        if offset is not None and offset.total_seconds():
            raise ValueError("Use DateTime64TZ for timezone-aware datetime")
        divisor = 10**self.precision
        value = value.astimezone(timezone.utc)
        ticks = int(value.timestamp() * divisor)
        Int64.write(buf, ticks)


class DateTime64TZ:
    precision = 3

    def __init__(self, precision):
        self.precision = precision

    @classmethod
    def __class_getitem__(cls, precision=3):
        return cls(precision)

    def read(self, buf: Buffer) -> datetime:
        ticks = Int64.read(buf)
        offset_minutes = Int16.read(buf)
        tz = timezone(timedelta(minutes=offset_minutes))

        divisor = 10**self.precision

        timestamp = ticks / divisor

        return datetime.fromtimestamp(timestamp, tz=tz)

    def write(self, buf: Buffer, value: datetime):

        divisor = 10**self.precision
        ticks = int(value.timestamp() * divisor)

        offset_td = value.utcoffset()
        if offset_td is None:
            raise ValueError("DateTime64TZ requires timezone-aware datetime")
        offset = int(offset_td.total_seconds() / 60)
        Int64.write(buf, ticks)
        Int16.write(buf, offset)


class Date:
    @classmethod
    def read(cls, buf: Buffer) -> date:
        return date(1970, 1, 1) + timedelta(days=buf.read_formated("H"))

    @classmethod
    def write(cls, buf: Buffer, value: date):
        days = (value - date(1970, 1, 1)).days
        if days < 0 or days > 2**16 - 1:
            raise ValueError("Date value out of range for Date type")
        buf.write_formated("H", (value - date(1970, 1, 1)).days)


T = TypeVar("T")


class Nullable(Generic[T]):
    def __init__(self, inner_type: TypeProto):
        self.inner_type = inner_type

    @classmethod
    def __class_getitem__(cls, inner_type: Type[TypeProto]):
        return cls(inner_type)

    def read(self, buf: Buffer) -> T | None:
        is_null = buf.read_formated("b")
        if is_null:
            return None

        return self.inner_type.read(buf)

    def write(self, buf: Buffer, value: T | None) -> None:
        if value is None:
            buf.write_formated("b", 1)
            return
        buf.write_formated("b", 0)
        self.inner_type.write(buf, value)



import typing as tp
from types import GenericAlias

import typing_extensions as te
from pydantic import BaseModel

from pybyntic.data_dumper import DataDumper
from pybyntic.data_parser import DataParser
from pybyntic.types import Skip


class AnnotatedBaseModel(BaseModel):
    @classmethod
    def deserialize(
        cls, data: bytes, decoder: tp.Optional[tp.Callable] = None
    ) -> te.Self:
        if decoder:
            data = decoder(data)
        parsed = {}
        parser = DataParser(data)
        for name, field in cls._get_types():
            value = parser.read(field)
            parsed[name] = value

        # Handle Skip fields - they should not be deserialized but need default values
        for name, model_field in cls.model_fields.items():
            if model_field.metadata and model_field.metadata[0] is Skip:
                if model_field.default is not None:
                    parsed[name] = model_field.default
                elif model_field.default_factory is not None:
                    parsed[name] = model_field.default_factory
                else:
                    parsed[name] = None
        parsed = cls._fix_nested(parsed)
        return cls(**parsed)

    def serialize(self, encoder: tp.Optional[tp.Callable] = None) -> bytes:
        dumper = DataDumper()
        for name, field in self._get_types():
            data = self._get_field(name)
            dumper.dump(field, data)

        value = dumper.get_value()
        if encoder:
            value = encoder(value)
        return value

    @classmethod
    def _get_types(cls) -> list[tuple[str, type[tp.Any]]]:
        model_types = []
        for name, model_field in cls.model_fields.items():
            annotation = model_field.annotation

            # Check for list-like types
            if isinstance(annotation, GenericAlias) and annotation.__origin__ is list:
                element_type = annotation.__args__[0]
                if isinstance(element_type, type) and issubclass(
                    element_type, AnnotatedBaseModel
                ):
                    for field_name, field_type in element_type._get_types():
                        model_types.append((name + "." + field_name, list[field_type]))  # type: ignore
                else:
                    model_types.append((name, model_field.metadata[0]))
            elif isinstance(annotation, type) and issubclass(
                annotation, AnnotatedBaseModel
            ):
                for field_name, field_type in annotation._get_types():
                    model_types.append((name + "." + field_name, field_type))
            elif model_field.metadata[0] is not Skip:
                model_types.append((name, model_field.metadata[0]))
        return model_types

    def _get_values(self) -> tp.Any:
        values = {}
        for name, field in self._get_types():
            values[name] = getattr(self, name)
        return values

    @classmethod
    def _normalize_lists(
        cls,
        node: tp.Any,
        path: tuple[str, ...] = (),
        *,
        zip_single_key: bool = False,  # set True if you DO want single-key zipping
    ) -> tp.Any:
        if isinstance(node, list):
            return [
                cls._normalize_lists(x, path, zip_single_key=zip_single_key)
                for x in node
            ]

        # Dicts: normalize children first
        if isinstance(node, dict):
            for k in list(node.keys()):
                node[k] = cls._normalize_lists(
                    node[k], path + (k,), zip_single_key=zip_single_key
                )

            # If all children are lists and (we allow single-key OR there are 2+ keys), zip into list[dict]
            if node:  # non-empty
                values = list(node.values())
                if all(isinstance(v, list) for v in values) and (
                    zip_single_key or len(node) > 1
                ):
                    lengths = {len(v) for v in values}
                    if len(lengths) != 1:
                        where = ".".join(path) or "<root>"
                        raise ValueError(
                            f"Arrays under '{where}' have mismatched lengths: {lengths}"
                        )
                    n = lengths.pop()

                    zipped: list[dict[str, tp.Any]] = []
                    # build entries by taking i-th element from each list
                    for i in range(n):
                        entry = {k: node[k][i] for k in node.keys()}
                        entry = cls._normalize_lists(
                            entry, path + (f"[{i}]",), zip_single_key=zip_single_key
                        )
                        zipped.append(entry)
                    return zipped

            return node

        return node

    @classmethod
    def _fix_nested(cls, original_dict: dict[str, tp.Any]) -> dict[str, tp.Any]:
        root: dict[str, tp.Any] = {}

        for key, value in original_dict.items():
            if "." not in key:
                # plain key lives at root level
                root[key] = value
                continue
            parts = key.split(".")
            d = root
            for p in parts[:-1]:
                d = d.setdefault(p, {})
                if not isinstance(d, dict):
                    # Safety: if something else already sits here, override with a dict
                    d = {}
            d[parts[-1]] = value
        # Recursively normalize: zip sibling lists, and descend into children
        return cls._normalize_lists(root)

    @staticmethod
    def _list_of_dicts_to_dict_of_lists(list_of_dicts):
        if not list_of_dicts:
            return {}

        # Initialize a dictionary to store lists of values
        dict_of_lists = {key: [] for key in list_of_dicts[0]}

        # Iterate over each dictionary in the list
        for d in list_of_dicts:
            for key, value in d.items():
                dict_of_lists[key].append(value)

        return dict_of_lists

    def _get_field(self, field_name: str) -> tp.Any:
        parts = field_name.split(".")

        def apply_part(value: tp.Any, part: str) -> tp.Any:
            """Apply a single path component `part` to `value` recursively."""
            # If we have a nested iterable, map recursively over its elements
            if isinstance(value, (list, tuple, set)):
                return type(value)(apply_part(v, part) for v in value)

            # Dict support (optional)
            if isinstance(value, dict):
                return value[part]

            # Otherwise treat as an object with attributes
            return getattr(value, part)

        current = self
        for p in parts:
            current = apply_part(current, p)
        return current



import struct
from typing import Any

import leb128  # type: ignore


# Thx to AsyncCH for the Buffer class
class Buffer:
    def __init__(self, data: bytes):
        self.buffer = bytearray(data)

    def _is_buffer_empty(self) -> bool:
        return len(self.buffer) == 0

    def is_buffer_readable(self) -> bool:
        if len(self.buffer) == 0:
            return False
        return True

    def _read_one(self) -> int:
        packet = self.buffer[:1]
        self.buffer = self.buffer[1:]
        return packet[0]

    def read_varint(self) -> int:
        packets = bytearray()
        while True:
            if not (self.is_buffer_readable()):
                break
            packet = self._read_one()
            packets.append(packet)
            if packet < 0x80:
                break
        return leb128.u.decode(packets)  # type: ignore

    def read_bytes(self, length: int) -> bytearray:
        packets = bytearray()

        while length > 0:
            if not (self.is_buffer_readable()):
                break
            packet = self.buffer[:1]
            packets.extend(packet)
            self.buffer = self.buffer[1:]
            length -= 1
        return packets

    def read_fixed_str(self, length: int, encoding: str = "utf-8") -> str:
        packet = self.read_bytes(length)
        return packet.decode(encoding=encoding)

    def read_str(self, encoding: str = "utf-8") -> str:
        length = self.read_varint()
        return self.read_fixed_str(length=length, encoding=encoding)

    def read_formated(self, fmt: str) -> Any:
        s = struct.Struct("<" + fmt)
        packet = self.read_bytes(s.size)
        return s.unpack(packet)[0]

    def write_bytes(self, data: bytes) -> None:
        self.buffer.extend(data)

    def write_varint(self, value: int) -> None:
        packets = bytearray()

        while True:
            byte = value & 0x7F
            value >>= 7
            if value != 0:
                byte |= 0x80
            packets.append(byte)
            if value == 0:
                break
        self.buffer.extend(packets)

    def write_formated(self, fmt: str, value: Any) -> None:
        b = struct.pack(fmt, value)
        self.write_bytes(b)
